//
// RoombaComm User Interface
//
//  Tod E. Kurt, tod@todbot.com
//
//

RoombaComm roombaComm = null;

// if you set this to not -1, it will not do the port picker UI
//int roombaCommPort = -1;
int roombaCommPort = 3;
int sensor_update_every = 5;  // update every 5 frames
// maps dir values to roomba radius numbers (absolute value)
int[] dirmap = { 1500, 1000, 500, 250, 125, 60, 30, 15 };
int max_dir  = dirmap.length - 1;
int max_speed = 5;
float angle_fudge_factor = 0.25;  // FIXME: why a divide-by-four?

int sensor_framecnt = 0;
int sdistance, sangle;
boolean safetyfault = false;
boolean sensorfault = false;

int c,d;
int octave = 3;

float rx,ry,rangle;

PFont fontA;
boolean toggleDraw = false;

// called by processing once on startup
void setup()
{
  size(350,350);
  stroke(100);
  background(153);
  ellipseMode(CENTER_RADIUS);
  rectMode(CENTER);
  smooth();
  framerate(10);

  fontA = loadFont("CourierNewPSMT-24.vlw"); 
  textFont(fontA, 12);
  fill(0);
  text("use arrow keys to drive",  10,height-28);

  roombaComm = new RoombaComm(this);
  if( roombaCommPort >= 0 ) {
    if( roombaComm.connect(roombaCommPort) ) {
      println("connected to Roomba!");
      roombaComm.startup();
    }
  }
  
  rx = width/2;
  ry = height/2;
}

// FIXME: put this in RoombaComm?
//
// Present the user with a serial port selector
//
void roombaCommSetup()
{
  int listlen = roombaComm.portlist.length;
  textFont(fontA, 12);
  background(153); fill(0); stroke(100);
  text("click a port:", 10,10);
  for( int i=0; i< listlen; i++ ) {
    text(i+": "+roombaComm.portlist[i], 10,20+10*i);
  }
  int s = (mouseY-15) / 10;
  if( s >= listlen || s < 0 ) return;
  rect(2,15+10*s, 10,8);
  if( mousePressed ) {
    line(10,20+10*s, width-10, 20+10*s);
    if( roombaComm.connect( s ) ) {
      roombaComm.startup();
      roombaCommPort = s;
    }
  }
}


void helpScreen()
{
  
}

void drawGridlines()
{
  int gdx = 50;
  stroke(100);    strokeWeight(1); 
  for( int i=0; i< 10; i++ ) {
    line(0, i*gdx, width, i*gdx);
    line(i*gdx,0, i*gdx, height);
  }
}

void drawRoomba()
{
  float radius = 20.0;

  if( toggleDraw ) {
//    image( roombapic, -radius,-radius);
  }
  else {
    noStroke();
    fill(45);
    ellipse(0, 0, radius, radius);
    fill(70);
    ellipse(0, 0, radius-3, radius-3);
    fill(160);
    rect(-radius+12,0, 15,4);
    fill(100);
    rect(radius/2,0, 4,4);
  }
}

// called by Processing 'framerate' times a second
void draw()
{
  if( roombaComm.connected == false ) { roombaCommSetup(); return; }

  textFont(fontA, 12);
  fill(0);  background(153); stroke(100);

  if( ++sensor_framecnt == sensor_update_every ) {
    sensorfault = ! roombaComm.updateSensors();
    sensor_framecnt=0;
    sdistance = roombaComm.distance();
    sangle    = roombaComm.angle();
    //sangle_dxeg= int((360 * sangle)/(258*PI));
    println("update at "+millis());
  }

  drawGridlines();

  // draw status panel
  text("use arrow keys to drive",  8,height-28);
  text("space to stop, / to reset",8,height-18);
  text("asdf.. music keyboard",    8,height-8);
  text("  status: "+ (sensorfault?"no roomba":(safetyfault?"safety fault":"OK")),8, 8);
  text("distance: "+sdistance, width-130, height-18);
  text("   angle: "+sangle,    width-130, height-8);

  // draw joystick
  strokeWeight(4);   // Thicker
  int jposx = width/2, jposy = height/2; 
  line( jposx, jposy, jposx - (d*5), jposy - (c*5) );
  
  float da = ((2.0 * sangle)/258.0) * angle_fudge_factor;
  float dx = sdistance * cos(da) / 10;
  float dy = sdistance * sin(da) / 10;
  rx = rx + dx;
  ry = ry + dy;
  rangle = rangle - da;
  rangle %= TWO_PI;
  println("sdist:"+sdistance+" sangle:"+sangle+" dx:"+dx+" dy:"+dy+" rangle:"+rangle);
  
  // torroidial mapping, sir                                                    
  if( rx > width ) rx = 0; if( rx < 0 ) rx = width;
  if( ry > width ) ry = 0; if( ry < 0 ) ry = height;

  if( roombaComm.safetyFault ) {
    textFont(fontA, 36);
    text("SAFETY FAULT!", 20,height/3);
  } 
  else {
    roombaComm.reset();
  }

  translate(rx,ry);
  rotate(rangle);
  drawRoomba();


    
}


void keyPressed()
{
  int note=-1;
  boolean movekey = false;
  if( key == CODED ) {
    if( keyCode == UP ) {
      c++; movekey=true;
    } else if( keyCode == DOWN ) {
      c--; movekey=true;
    } else if( keyCode == LEFT ) {
      d++; movekey=true;
    } else if( keyCode == RIGHT ) {
      d--; movekey=true;
    }
  }
  else if( key == ' ' ) {
    roombaComm.stop();
    c = 0; d = 0; 
  }
  else if( key == '/' ) {
    roombaComm.reset();  // in case we pick the roomba up and it shifts to passive mode
    c = 0; d = 0;
  }
  else {
    switch( key ) {
      // pseudo-keyboard to play notes on
      case 'a': note = 0;  break; 
      case 'w': note = 1;  break; 
      case 's': note = 2;  break; 
      case 'e': note = 3;  break; 
      case 'd': note = 4;  break; 
      case 'f': note = 5;  break; 
      case 't': note = 6;  break; 
      case 'g': note = 7;  break; 
      case 'y': note = 8;  break; 
      case 'h': note = 9;  break; 
      case 'u': note =10;  break; 
      case 'j': note =11;  break; 
      case 'k': note =12;  break; 
      case 'o': note =13;  break; 
      case 'l': note =14;  break; 
      case 'z': octave--;  break;  // change octaves
      case 'x': octave++;  break;  // change octaves
    }
  }
  
  // we actually hit a note key
  if( note >= 0 ) {
    octave = (octave < 1) ? 1 : (octave > 7) ? 7 : octave;
    roombaComm.playNote( note+(octave*12) + 36, 10);
  }

  if( movekey ) {
    if( c>max_speed ) c=max_speed; if( c<-max_speed ) c=-max_speed;
    if( d>max_dir )   d=max_dir;   if( d<-max_dir )   d=-max_dir;
    
    println("c:"+c+", d:"+d);
    
    if( c == 0 && d == 0 ) {
      roombaComm.stop();
    } else if( d==0 ) {
      roombaComm.goStraight( c*75 );
    } else if( c==0 ) {
      if( d > 0 )
        roombaComm.spinLeft( d*50 );
      else
        roombaComm.spinRight( -d*50 );
    } else {
      if( d>0 )
        roombaComm.drive( c*75, dirmap[d] );
      else 
        roombaComm.drive( c*75, -dirmap[-d] );
    }
  }
  
  
}
