//
// RoombaComm User Interface
//
//  Tod E. Kurt, tod@todbot.com
//
//

RoombaComm roombaComm = null;

// if you set roombaCommPort to -1, it will do the port picker UI
int roombaCommPort = -1;
//int roombaCommPort = 3;
//int roombaCommPort = 8;
int myframerate = 15;
int sensor_update_every = 5;  // update every 5 frames

// map dir values to roomba radius numbers (absolute value)
int[] dirmap = { 1500, 1000, 500, 250, 125, 60, 30, 15 };
int max_dir  = dirmap.length ;
int [] speedmap = { 100, 200, 300, 400, 500 };
// int [] speedmap = { -500, -400, -300, -200, -100, 0, 100, 200, 300, 400, 500 };
int max_speed = speedmap.length;

float angle_fudge_factor = 0.25;  // FIXME: why a divide-by-four?
float dist_fudge_factor = 0.25;

int sensor_framecnt = 0;
int sdistance, sangle;
boolean sensorfault = true;
boolean safetyfault = true;

int c,d;
int octave = 3;  // which octave we're on when playing notes
boolean vacuumOn = false;

float rx,ry,rangle;  // display info for displayed roomba icon

PFont fontA;
PImage rpic;
boolean toggleDraw = false;
float rradius = 20.0;


// called by processing once on startup
void setup()
{
  size(450,450);
  background(153); stroke(100); 
  ellipseMode(CENTER_RADIUS);
  rectMode(CENTER);
  smooth();
  framerate(myframerate);

  rpic = loadImage("roomba-sm.png");

  fontA = loadFont("CourierNewPSMT-24.vlw"); 
  textFont(fontA, 12);
  fill(0);

  roombaComm = new RoombaComm(this);
  if( roombaCommPort >= 0 ) {
    if( roombaComm.connect(roombaCommPort) ) {
      println("connected to Roomba!");
      roombaComm.startup();
    }
  }
  
  rx = width/2;
  ry = height/2;
  rangle = 0;
}

// called by Processing 'framerate' times a second
void draw()
{
  if( roombaComm.connected == false ) { roombaCommSetup(); return; }

  textFont(fontA, 12);
  fill(0);  background(153); stroke(100);

  updateRoombaState();
  
  drawGridlines();
  drawJoystick(); 
  drawStatus();

  drawRoomba();
}

void updateRoombaState() {
    if( ++sensor_framecnt == sensor_update_every ) {
    sensorfault = ! roombaComm.updateSensors();
    if( ! roombaComm.safetyFault ) {
      if( safetyfault ) {
        roombaComm.reset();
        c = d = 0;
        safetyfault = false;
      } 
    } else {
      safetyfault = true;
    }
    sensor_framecnt=0;
    sdistance = roombaComm.distance();
    sangle    = roombaComm.angle();
    //sangle_dxeg= int((360 * sangle)/(258*PI));
  }
  
  // compute drawing params from roomba sensor data
  // FIXME: for some reason this is screwing up when roomba goes into reverse
  float da = ((2.0 * sangle)/258.0) * angle_fudge_factor;
  rangle = rangle - da;
  rangle %= TWO_PI;
  float dx = sdistance * sin(rangle) * (0.05);
  float dy = sdistance * cos(rangle) * (0.05);
  rx = rx + dx;
  ry = ry - dy;
  //println("sdist:"+sdistance+" sangle:"+sangle+" dx:"+dx+" dy:"+dy+" rangle:"+rangle);
  // torroidial mapping, sir                                                    
  if( rx > width  ) rx = 0; if( rx < 0 ) rx = width;
  if( ry > height ) ry = 0; if( ry < 0 ) ry = height;

}

// called by Processing whenever a key is pressed
// called repeatedly if a key is held down
void keyPressed()
{
  int note=-1;
  boolean movekey = false;
  if( key == CODED ) {
    if( keyCode == UP ) {
      c++; movekey=true;
    } else if( keyCode == DOWN ) {
      c--; movekey=true;
    } else if( keyCode == LEFT ) {
      d++; movekey=true;
    } else if( keyCode == RIGHT ) {
      d--; movekey=true;
    }
  }
  else if( key == ' ' ) {
    roombaComm.stop();
    c = 0; d = 0;
  }
  else if( key == '/' ) {
    roombaComm.reset();  // in case we pick the roomba up and it shifts to passive mode
    c = 0; d = 0;
    rx = width/2; ry = height/2; rangle = 0;
  }
  else if( key == ',' ) {
    toggleDraw = !toggleDraw;
  }
  else if( key == '\t' ) {
      vacuumOn = !vacuumOn;
      roombaComm.vacuum( vacuumOn );
  }
  else {
    switch( key ) {
      // pseudo-keyboard to play notes on
      case 'a': note = 0;  break; 
      case 'w': note = 1;  break; 
      case 's': note = 2;  break; 
      case 'e': note = 3;  break; 
      case 'd': note = 4;  break; 
      case 'f': note = 5;  break; 
      case 't': note = 6;  break; 
      case 'g': note = 7;  break; 
      case 'y': note = 8;  break; 
      case 'h': note = 9;  break; 
      case 'u': note =10;  break; 
      case 'j': note =11;  break; 
      case 'k': note =12;  break; 
      case 'o': note =13;  break; 
      case 'l': note =14;  break; 
      case 'z': octave--;  break;  // change octaves
      case 'x': octave++;  break;  // change octaves
    }
  }
  
  // we actually hit a note key
  if( note >= 0 ) {
    octave = (octave < 1) ? 1 : (octave > 7) ? 7 : octave;
    roombaComm.playNote( note+(octave*12) + 36, 10);
  }

  if( movekey ) {
    if( c>max_speed ) c=max_speed; if( c<-max_speed ) c=-max_speed;
    if( d>max_dir )   d=max_dir;   if( d<-max_dir )   d=-max_dir;
    
    int ca = abs(c);  int cs = (c>0)?1:(c<0)?-1:0;  // absolute value & sign
    int sp = c*100; // c!=0 ? cs * speedmap[ca-1];
    if( c == 0 && d == 0 ) {
      roombaComm.stop();
    } else if( d==0 ) {
      roombaComm.goStraight( sp );
    } else if( c==0 ) {
      if( d > 0 )
        roombaComm.spinLeft( d*50 );
      else
        roombaComm.spinRight( -d*50 );
    } else {
      if( d>0 )
        roombaComm.drive( sp, dirmap[d-1] );
      else 
        roombaComm.drive( sp, -dirmap[-d-1] );
    }
  }  
}

// nice little grid to help judge onscreen motion
void drawGridlines()
{
  int gridspacing = 50;
  stroke(120);    strokeWeight(1); 
  for( int i=0; i< 10; i++ ) {
    line(0, i*gridspacing, width, i*gridspacing);
    line(i*gridspacing,0, i*gridspacing, height);
  }
}

// little roomba icon or scanned picture
void drawRoomba()
{
  translate(rx,ry);
  rotate(rangle);

  if( toggleDraw ) {
    image( rpic, -rradius,-rradius);
  }
  else {
    noStroke();
    fill(45);
    ellipse(0, 0, rradius, rradius);
    fill(70);
    ellipse(0, 0, rradius-3, rradius-3);
    fill(160);
    rect(0,-rradius+12, 4,15);
    fill(100);
    rect(0,rradius/2, 4,4);
  }
}

// joystiq indicating current speed/direction controls
void drawJoystick()
{
  int jposx0 = width-50, jposy0 = 150; 
  int jposx1 = int(jposx0 - (d*5)), jposy1 = int(jposy0 - (c*5));
  fill(160);  noStroke();
  rect(jposx0,jposy0, 50,50);
  stroke(100); strokeWeight(4);   
  line( jposx0, jposy0, jposx1, jposy1 );  
  stroke(200,100,100); strokeWeight(8);
  line( jposx1, jposy1, jposx1, jposy1 );  
  //  line( jposx, jposy, jposx - (d*5), jposy - (c*5) );
}

// status messages  (FIXME: make all these pixel positions relative)
void drawStatus()
{ 
  textFont(fontA, 10);
  fill(160);  noStroke();
  rect(width-50,40, 90,70);  // sensor background
  rect(width-50,100, 60,40);  // buttons background
  
  fill(0); stroke(100); strokeWeight(5);
  
  text("cliff", width-63, 14);
  if( roombaComm.cliffLeft()       ) stroke(255,0,0); else stroke(100);
  line(width-90,20,width-87,12);  
  if( roombaComm.cliffFrontLeft()  ) stroke(255,0,0); else stroke(100);
  line(width-80,12,width-70,12);
  if( roombaComm.cliffFrontRight() ) stroke(255,0,0); else stroke(100);
  line(width-30,12,width-20,12);
  if( roombaComm.cliffRight()      ) stroke(255,0,0); else stroke(100);
  line(width-10,20,width-13,12);

  text("bump", width-60, 24);
  if( roombaComm.bumpLeft()  ) stroke(255,0,0); else stroke(100);
  line(width-80,25,width-70,22);
  if( roombaComm.bumpRight() ) stroke(255,0,0); else stroke(100);
  line(width-30,22,width-20,25);

  text("wheels", width-67, 44);  
  if( roombaComm.wheelDropLeft()  ) stroke(255,0,0); else stroke(100);
  line(width-80,42,width-70,42);
  if( roombaComm.wheelDropRight() ) stroke(255,0,0); else stroke(100);
  line(width-30,42,width-20,42);
  if( roombaComm.wheelDropCenter()) stroke(255,0,0); else stroke(100);
  line(width-52,32,width-48,32);

  text("dirt", width-60,54);
  boolean dirt = roombaComm.dirt();
  if( dirt ) stroke(0,0,200); else stroke(100);
  line(width-80,52,width-70,52); 
  line(width-30,52,width-20,52); 

  text("p", width-70,100);
  if( roombaComm.powerButton() ) stroke(0,200,0); else stroke(100);
  line(width-70,105, width-70+2,105);
  text("s", width-50,100);
  if( roombaComm.spotButton()  ) stroke(0,200,0); else stroke(100);
  line(width-50,105, width-50+2,105);
  text("c", width-40,100);
  if( roombaComm.cleanButton() ) stroke(0,200,0); else stroke(100);
  line(width-40,105, width-40+2,105);
  text("m", width-30,100);
  if( roombaComm.maxButton()   ) stroke(0,200,0); else stroke(100);
  line(width-30,105, width-30+2,105);

  textFont(fontA, 14);
  text("sensors",       width-77,70 );
  text("buttons",       width-77,90 );

  text("status: "+ (sensorfault?"no roomba":(safetyfault?"safety fault":"OK")), 8,12);

  textFont(fontA, 12);      
  text("use arrow keys to drive",               10,height-48);
  text("space to stop, / to reset roomba",      10,height-38);
  text("comma toggles view, tab toggles vacuum",10,height-28);
  text("asdf... plays notes",                   10,height-18);
  
  text("distance: "+sdistance, width-130, height-18);
  text("   angle: "+sangle,    width-130, height-8);
//  text("angle:"+int(angle*360/TWO_PI), width-90,height-18);
//  text("speed:"+speed, width-90,height-8);

  if( safetyfault ) {
    println("safteyfault!");
    textFont(fontA, 36);
    text("SAFETY FAULT!", 30,height/3);
    textFont(fontA, 24);
    text("reposition Roomba to reset", 40,height/3+20);
  }

}




// FIXME: put this in RoombaComm?
//
// Present the user with a serial port selector
//
void roombaCommSetup()
{
  int listlen = roombaComm.portlist.length;
  textFont(fontA, 12);
  background(153); fill(0); stroke(100);
  text("click a port:", 10,10);
  for( int i=0; i< listlen; i++ ) {
    text(i+": "+roombaComm.portlist[i], 10,20+10*i);
  }
  int s = (mouseY-15) / 10;
  if( s >= listlen || s < 0 ) return;
  rect(2,15+10*s, 10,8);
  if( mousePressed ) {
    line(10,20+10*s, width-10, 20+10*s);
    if( roombaComm.connect( s ) ) {
      roombaComm.startup();
      roombaCommPort = s;
    }
  }
}

